#include <stdio.h>
#include <stdlib.h>

#include <xf86Parser.h>
#include <xf86Optrec.h>
#include <X11/Intrinsic.h>

XF86ConfScreenPtr find_screen(XF86ConfigPtr conf)
{
	XF86ConfLayoutPtr layout=conf->conf_layout_lst;
	if (!layout) return NULL;
	XF86ConfAdjacencyPtr adjacency = layout->lay_adjacency_lst;
	if (!adjacency) return NULL;
	return adjacency->adj_screen;
}

XF86ConfDevicePtr find_device(XF86ConfigPtr conf)
{
	/* go to the way layout --> screen -> device */
	XF86ConfScreenPtr screen=find_screen(conf);
	if (!screen) return NULL;
	return screen->scrn_device;
}


XF86ConfInputPtr find_input(XF86ConfigPtr conf,char *type)
{
	/* go to the way layout --> first input refer "CorePointer or CoreKeyboard" */
	XF86ConfLayoutPtr layout=conf->conf_layout_lst;
	XF86ConfInputrefPtr inputref = layout->lay_input_lst;
	for(inputref = layout->lay_input_lst;inputref;inputref=inputref->list.next)
	{
		XF86OptionPtr  p;
		for (p = inputref->iref_option_lst;p;p=p->list.next)
			if (!strcmp(p->opt_name,type))
			return inputref->iref_inputdev;
	}
	return 0;
}

XF86ConfMonitorPtr find_monitor(XF86ConfigPtr conf)
{
	/* go to the way layout --> screen --> monitor */
	return find_screen(conf)->scrn_monitor;
}


int xf86renameMonitor(XF86ConfigPtr config, XF86ConfMonitorPtr mon_list, const char *name)
{
    int num = 0;

    if (config == NULL || mon_list == NULL || name == NULL || *name == '\0')
	return (False);


    while (mon_list != NULL)
    {
	char * name_numbered;
	XF86ConfScreenPtr scr = config->conf_screen_lst;
	name_numbered = malloc(strlen(name) + 3);
	sprintf(name_numbered, "%s|%d",name,num);
	while (scr != NULL) {
	    if (scr->scrn_monitor == mon_list) {
		XtFree(scr->scrn_monitor_str);
		scr->scrn_monitor_str = XtNewString(name_numbered);
	    }

	    scr = (XF86ConfScreenPtr)(scr->list.next);
	}
	XtFree(mon_list->mon_identifier);
	mon_list->mon_identifier = XtNewString(name_numbered);
	mon_list = mon_list->list.next;
	++num;
    }
    return (True);
}


int
xf86renameDevice(XF86ConfigPtr config, XF86ConfDevicePtr dev,const char *name)
{
    int num = 0;
    XF86ConfDevicePtr dev_first = dev;
    if (config == NULL || dev == NULL || name == NULL || *name == '\0')
	return (False);
    while (dev)
    {
	char * name_numbered;
	XF86ConfScreenPtr scr = config->conf_screen_lst;
	name_numbered = malloc(strlen(name) + 3);
	sprintf(name_numbered, "%s|%d",name,num);
	while (scr != NULL) {
	    if (scr->scrn_device == dev) {
		XtFree(scr->scrn_device_str);
		scr->scrn_device_str = XtNewString(name_numbered);
	    }

	    scr = (XF86ConfScreenPtr)(scr->list.next);
	}

	XtFree(dev->dev_identifier);
	dev->dev_identifier = XtNewString(name_numbered);
	dev = dev->list.next;
	++num;
    }
    dev = dev_first;
    return (True);
}

int xf86renameInput(XF86ConfigPtr config, XF86ConfInputPtr input, const char *name)
{
    XF86ConfLayoutPtr lay = config->conf_layout_lst;

    if (config == NULL || input == NULL || name == NULL || *name == '\0')
	return (False);

    while (lay != NULL) {
	XF86ConfInputrefPtr iref = lay->lay_input_lst;

	while (iref != NULL) {
	    if (strcasecmp(input->inp_identifier, iref->iref_inputdev_str) == 0) {
		XtFree(iref->iref_inputdev_str);
		iref->iref_inputdev_str = XtNewString(name);
	    }
	    iref = (XF86ConfInputrefPtr)(iref->list.next);
	}
	lay = (XF86ConfLayoutPtr)(lay->list.next);
    }

    XtFree(input->inp_identifier);
    input->inp_identifier = XtNewString(name);

    return (True);
}
