#include <getopt.h>

#include <stdio.h>
#include <stdlib.h>

#include <xf86Parser.h>
#include <xf86Optrec.h>
#include <X11/Intrinsic.h>

extern int xf86renameDevice(XF86ConfigPtr config, XF86ConfDevicePtr dev,const char *name);
int xf86renameInput(XF86ConfigPtr config, XF86ConfInputPtr input, const char *name);
int xf86renameMonitor(XF86ConfigPtr config, XF86ConfMonitorPtr mon, const char *name);

extern XF86ConfDevicePtr find_device(XF86ConfigPtr conf);
extern XF86ConfMonitorPtr find_monitor(XF86ConfigPtr conf);
extern XF86ConfScreenPtr find_screen(XF86ConfigPtr conf);

extern const char *__progname;

void extract_option(char *str,char **name,char **value)
{
  char *quote1=str,*quote2,*quote3,*quote4;
	
  if(*quote1 != '"') return;
  quote2 = index(quote1+1,'"');
  if (!quote2) return;
  quote3 = index(quote2+1,'"');
  if (quote3)
    {
      quote4=index(quote3+1,'"');
      if (!quote4) return;
    }

  *name=quote1+1;
  *quote2=0;

  if (quote3)
    {
      *value=quote3+1;
      *quote4=0;
    }
}


#define CLEAR_OPTIONS(x)			\
  xf86optionListFree(p->x);			\
  p->x = 0;

#define CLEAR_COMMENT(x)			\
  XtFree(p->x);					\
  p->x=XtNewString("");

#define MergeLists(type, head, tail) \
    head = (type) xf86addListItem((GenericListPtr) (head), (GenericListPtr) (tail))

#define AddListItem(type, head, new) \
    (new)->list.next = NULL; \
    MergeLists(type, head, new)

#define LoadList conf->conf_modules->mod_load_lst

void get_loads(XF86ConfigPtr conf)
{
  XF86LoadPtr l;

  for (l = LoadList; l != NULL; l = l->list.next)
    if (l->load_type == XF86_LOAD_MODULE)
      printf("%s\n", l->load_name);
}

#ifdef XF86_DISABLE_MODULE
#define DisableList conf->conf_modules->mod_disable_lst

void inline set_module(XF86ConfigPtr conf, const char *module, int in, int out)
{
  XF86LoadPtr l;
  int f = 0;

  if (LoadList != DisableList)
    MergeLists(XF86LoadPtr, LoadList, DisableList);
  for (l = LoadList; l != NULL; l = l->list.next)
    if ((l->load_opt == NULL) && (strcmp(l->load_name, module) == 0))
      {
        if (l->load_type == in)
          {
            l->load_type = out;
            l->load_comment = NULL;
            l->ignore = 0;
          }
        f = 1;
      }
  if ((f == 0) && ((l = malloc(sizeof(XF86LoadRec))) != NULL))
    {
      l->load_name = (char*) module;
      l->load_opt = NULL;
      l->load_type = out;
      l->load_comment = NULL;
      l->ignore = 0;
      AddListItem(XF86LoadPtr, LoadList, l);
    }
}

void set_loads(XF86ConfigPtr conf, const char *module)
{
  set_module(conf, module, XF86_DISABLE_MODULE, XF86_LOAD_MODULE);
}

void get_disabled(XF86ConfigPtr conf)
{
  XF86LoadPtr l;

  for (l = DisableList; l != NULL; l = l->list.next)
      if (l->load_type == XF86_DISABLE_MODULE)
        printf("%s\n", l->load_name);
}

void set_disable(XF86ConfigPtr conf, const char *module)
{
  set_module(conf, module, XF86_LOAD_MODULE, XF86_DISABLE_MODULE);
}
#endif

#define ExtOptList conf->conf_extensions->ext_option_lst

void get_extopts(XF86ConfigPtr conf)
{
  if (conf->conf_extensions != NULL)
    {
      XF86OptionPtr p;
      for (p = ExtOptList; p != NULL; p = xf86nextOption(p))
        printf("%s=%s\n", xf86optionName(p), xf86optionValue(p));
    }
}

void get_extopt(XF86ConfigPtr conf, const char *name)
{
  if (conf->conf_extensions != NULL)
    {
      XF86OptionPtr p;
      if ((p = xf86findOption(ExtOptList, name)) != NULL)
        printf("%s\n", xf86optionValue(p));
    }
}

void set_extopt(XF86ConfigPtr conf, char *opt)
{
  char *c = strchr(opt, '=');

  if (c != NULL)
    {
      if (conf->conf_extensions == NULL)
        {
          conf->conf_extensions = malloc(sizeof(XF86ConfExtensionsRec));
          if (conf->conf_extensions == NULL) return;
          ExtOptList = NULL;
          conf->conf_extensions->extensions_comment = NULL;
        }
      *c = '\0';
      ExtOptList = xf86addNewOption(ExtOptList, opt, &(c[1]));
    }
}

void setup_monitor(XF86ConfigPtr conf,const char *monitor_name)
{
  char * line = NULL;
  size_t len = 0;
  XF86ConfMonitorPtr p = find_monitor(conf);
  XF86ConfMonitorPtr p_first = p;
  if (!p) return;
	
  xf86renameMonitor(conf,p,monitor_name);

  //reset some values to zero state
  while (p)
    {
      p->mon_n_hsync = 0;
      p->mon_n_vrefresh = 0;
      p->mon_width = 0;
      p->mon_height = 0;
      p = p->list.next;
    }
  p = p_first;
  //clear all previous options and comments
  while (p)
    {
      CLEAR_OPTIONS(mon_option_lst)
	CLEAR_COMMENT(mon_comment)
	p = p->list.next;
    }
  p = p_first;
  while(getline(&line,&len,stdin) != -1)
    {
      char *sep = index(line,':');
      if (!sep) continue;

      char *name=line;
      char *value=sep+1;
      *sep=0;
		
      if (strlen(value) <= 1) continue; //ignore empty strings ( with single '\n' character)
		
      if (!strcmp(name,"width"))
	{
	  int width;
	  int res = sscanf(value,"%d",&width);
	  if (res == 1)
	    while (p)
	      {
		p->mon_width = width;
		p = p->list.next;
	      }
	  p = p_first;
	}
      else if (!strcmp(name,"height"))
	{
	  int height;
	  int res = sscanf(value,"%d",&height);
	  if (res == 1)
	    while (p)
	      {
		p->mon_height = height;
		p = p->list.next;
	      }
	  p = p_first;
	}
      else if (!strcmp(name,"hsync") && strcmp(value,""))
	{
	  float losync,hisync;
	  int res = sscanf(value,"%f-%f",&losync,&hisync);
	  if (res == 1) { hisync=losync; res = 2; }//single frequency was received
	  if (res == 2)
	    while (p)
	      {
		p->mon_hsync[0].hi = hisync;
		p->mon_hsync[0].lo = losync;
		p->mon_n_hsync = 1;
		p = p->list.next;
	      }
	  p = p_first;
	}
      else if (!strcmp(name,"vsync") && strcmp(value,""))
	{
	  float losync,hisync;
	  int res = sscanf(value,"%f-%f",&losync,&hisync);
	  if (res == 1) { hisync=losync; res = 2; }//single frequency was received
	  if (res == 2)
	    while (p)
	      {
		p->mon_vrefresh[0].hi = hisync;
		p->mon_vrefresh[0].lo = losync;
		p->mon_n_vrefresh = 1;
		p = p->list.next;
	      }
	  p = p_first;
	}
    }	
}


void setup_input(XF86ConfigPtr conf,XF86ConfInputPtr p,const char *input_name)
{
  char * line = NULL;
  size_t len = 0;
	
  xf86renameInput(conf,p,input_name);

  //clear all previous options and comments
  CLEAR_OPTIONS(inp_option_lst)
    CLEAR_COMMENT(inp_comment)


    while(getline(&line,&len,stdin) != -1)
      {
	char *sep = index(line,':');

	if (!sep) continue;
		
	char *name = line;
	char *value = sep+1;
	char *pos = sep-1;

	while(*pos == ' ' || *pos == '\t') --pos;
	while(*name == ' ' || *name == '\t') ++name;
	while(*value == ' ' || *value == '\t') ++value;

	value[strlen(value)-1]=0;//remove "\n"
	*(pos+1)=0;

	if (!strcmp(name,"X11 Driver"))
	  {
	    XtFree(p->inp_driver);
	    p->inp_driver=XtNewString(value);
	  }
	else if (!strcmp(name,"X11 config"))
	  {
	    char *confname = value;
	    char *pos = index(confname,' ');
	    if (!pos) pos = index(confname,'\t');
	    if (pos)
	      {
		char *confcontent = pos+1;
		*pos = 0;
		if (!strcmp(confname,"Option"))
		  {
		    char *opt_name;
		    char *opt_value;
		    extract_option(confcontent,&opt_name,&opt_value);
		    p->inp_option_lst = xf86addNewOption(p->inp_option_lst,
							 XtNewString(opt_name),
							 XtNewString(opt_value));
		  }
	      }
	  }

      }
}

void setup_device(XF86ConfigPtr conf,const char *dev_name)
{
  int was_driver = 0;
  char * line = NULL;
  size_t len = 0;
  XF86ConfDevicePtr p = find_device(conf);
  XF86ConfDevicePtr p_first = p;
  if (!p) return;

  xf86renameDevice(conf,p,dev_name);

  //clear all previous options and comments
  for(p = p_first;p != NULL;p = p->list.next)
    {
      CLEAR_OPTIONS(dev_option_lst)
	CLEAR_COMMENT(dev_comment)
	}

  while(getline(&line,&len,stdin) != -1)
    {
      char *pos = index(line,'\t');
      if (!pos) continue;
		
      char *value=pos+1;
      char *name=line;

      value[strlen(value)-1]=0;//remove "\n"
      *pos = 0;//made fake string
      while (value && (*value == ' ' || *value == '\t')) ++value;

      if (!strcmp(name,"xdriver") && !was_driver)
	{
	  for(p = p_first;p != NULL;p = p->list.next)
	    {
	      XtFree(p->dev_driver);
	      p->dev_driver=XtNewString(value);
	      was_driver = 1;
	    }

	}
      else if (!strcmp(name,"config"))
	{
	  for(p = p_first;p != NULL;p = p->list.next)
	    {
	      if (value[0] == '#')
		p->dev_comment = xf86addComment(p->dev_comment,value);
	      else
		{
		  char *opt_name = value;
		  char *pos = index(value,' ');
		  if (!pos) pos = index(value,'\t');
		  if (pos) *(pos++) = 0;
		  while (pos && (*pos == ' ' || *pos == '\t')) ++pos;
		  if (pos && (!strcmp(opt_name,"MemBase")))
		    sscanf(pos,"%lx",&(p->dev_mem_base));
		  else if (pos && (!strcmp(opt_name,"Option")))
		    {
		      char *comment = index(pos,'#'); //find and strip comment
		      if (comment) *(comment++)=0;

		      char *opt_name=0,*opt_value=0;
		      extract_option(pos,&opt_name,&opt_value);
		      p->dev_option_lst = xf86addNewOption(p->dev_option_lst,
							   XtNewString(opt_name),
							   (opt_value)?XtNewString(opt_value):0);
		      if (comment)
			{
			  XF86OptionPtr o= xf86findOption(p->dev_option_lst,opt_name);
			  if (o) o->opt_comment = xf86addComment(o->opt_comment,comment);
			}
		    }
		}
	    }
	}
    }
}


void clear_modes(XF86ModePtr ptr)
{
  XF86ModePtr prev;
  while (ptr)
    {
      XtFree(ptr->mode_name);
      prev = ptr;
      ptr = ptr->list.next;
      free(prev);
    }
}


void setup_resolution(XF86ConfigPtr conf,const char *res_list)
{
  XF86ConfScreenPtr screen = find_screen(conf);
  XF86ConfDisplayPtr disp;

  for(screen = find_screen(conf); screen; screen = screen->list.next)
    {
      //first clear all old modes
      for(disp = screen->scrn_display_lst; disp; disp = disp->list.next)
	{
	  clear_modes(disp->disp_mode_lst);
	  disp->disp_mode_lst = 0;
	}
      if (!strcmp(res_list,"auto")) continue;
      //for each founded resolution prepend to list
      char *res_parse=strdup(res_list);
      while(res_parse)
	{
	  char *next_pos = index(res_parse,',');
	  if (next_pos) *(next_pos++)=0;
	  //cons for each display entry
	  for(disp = screen->scrn_display_lst; disp; disp = disp->list.next)
	    {
	      XF86ModePtr mode = calloc(1, sizeof(XF86ModeRec));
	      mode->mode_name = XtNewString(res_parse);
	      AddListItem(XF86ModePtr, disp->disp_mode_lst, mode);
	    }
		
	  res_parse = next_pos;
	}
      free(res_parse);
    }
}

void
usage (int retcode)
{
  fprintf (stdout,
	   "Usage: %s <action> [<action>...] old_config [new_config] \n"
	   "\tnew_config is necessarily for all modifications",
	   __progname);
  fputs ("\nTree - utility to read and modify xorg config file\n",
	 stdout);

  fputs ("\t-h, --help		display help screen\n"
	 "\t-v, --version		display version information\n"
	 "\t-d, --get-device	get current device identifier\n"
	 "\t-D, --set-device	set current device parameters\n"
	 "\t-n, --get-monitor	get current monitor identifier\n"
	 "\t-N, --set-monitor	set current monitor parameters\n"
	 "\t-c, --get-color-depth	get current default color depth value\n"
	 "\t-C, --set-color-depth	set new default color depth value\n"
	 "\t-r, --get-resolution	get current resolution list (mode list)\n"
	 "\t-R, --set-resolution	set new resolution list (mode list)\n"
	 "\t-l, --get-loads		get current modules list to load\n"
#ifdef XF86_DISABLE_MODULE
	 "\t-L, --set-loads		set new module to load\n"
	 "\t-m, --get-disable	get current modules list to disable\n"
	 "\t-M, --set-disable	set module to disable\n"
#endif
	 "\t-a, --get-extopts	get options from Section Extensions (<name>=<value>)\n"
	 "\t-E, --set-extopt	set option in Section Extensions (<name>=<value>)\n"
	 "\t-e, --get-extopt	get specified option from Section Extensions\n",
	 stdout);

  fputs ("\nReport bugs to https://bugzilla.altlinux.org\n", stdout);
  exit (retcode);
}

typedef
enum
  {gdevice,
   sdevice,
   gmon,
   smon,
   gcolor,
   scolor,
   gres,
   sres,
   gdriver,
   gloads,
#ifdef XF86_DISABLE_MODULE
   sloads,
   gdisable,
   sdisable,
#endif
   gextopts,
   sextopt,
   gextopt} action_type;

typedef struct ActionRec
  {
    GenericListRec list;
    action_type type;
    char *arg;
  } ActionRec, *ActionPtr;

ActionPtr add_action_to_list(ActionPtr actions, action_type a, const char *arg)
{
  ActionPtr action = malloc(sizeof(ActionRec));

  if (action != NULL)
    {
      action->type = a;
      action->arg = (char*) arg;
      AddListItem(ActionPtr, actions, action);
      return actions;
    }
  else
    return actions;
}

#define add_action(a, arg) add_action_to_list(actions, (a), (arg))
#define add_info(a) add_action((a), NULL)

int main(int argc,char *argv[])
{
  const char *filename;
  XF86ConfigPtr conf;
  char *oldconfig = 0;
  char *newconfig = 0;
  ActionPtr actions = NULL;
  int write_config = 0;

  while (1)
    {
      static struct option long_options[] = {
	{"help", no_argument, 0, 'h'},
	{"version", no_argument, 0, 'v'},
	/*visualisation */
	{"get-device", no_argument, 0, 'd'},
	{"set-device", required_argument, 0, 'D'},
	{"get-monitor", no_argument, 0, 'n'},
	{"set-monitor", required_argument, 0, 'N'},
	{"get-color-depth", no_argument, 0, 'c'},
	{"set-color-depth", required_argument, 0, 'C'},
	{"get-resolution", no_argument, 0, 'r'},
	{"set-resolution", required_argument, 0, 'R'},
	{"get-loads", no_argument, 0, 'l'},
#ifdef XF86_DISABLE_MODULE
	{"set-loads", required_argument, 0, 'L'},
	{"get-disable",no_argument, 0, 'm'},
	{"set-disable",required_argument, 0, 'M'},
#endif
	{"get-extopts",no_argument, 0, 'a'},
	{"set-extopt", required_argument, 0, 'E'},
	{"get-extopt", required_argument, 0, 'e'},
	{0, 0, 0, 0}
      };
      char c;

#ifdef XF86_DISABLE_MODULE
      c = getopt_long (argc, argv, "hvdD:nN:cC:rR:lL:mM:a:eE:", long_options, NULL);
#else
      c = getopt_long (argc, argv, "hvdD:nN:cC:rR:la:eE:", long_options, NULL);
#endif
      if (c == -1)
	break;
      switch (c)
	{
	case 'h':
	  usage (EXIT_SUCCESS);
	case 'v':
	  fputs ("XConfig reader/changer " VERSION "\n"
		 "Written by Stanislav Ievlev\n"
		 "Completed by Led\n\n"
		 "Copyright (C) 2005,2008 ALT Linux Team\n",
		 stdout);
	  exit (EXIT_SUCCESS);
	case 'd':
	  actions = add_info(gdevice);
	  break;
	case 'D':
	  actions = add_action(sdevice, strdup(optarg));
	  write_config++;
	  break;
	case 'n':
	  actions = add_info(gmon);
	  break;
	case 'N':
	  actions = add_action(smon, strdup(optarg));
	  write_config++;
	  break;
	case 'c':
	  actions = add_info(gcolor);
	  break;
	case 'C':
	  actions = add_action(scolor, strdup(optarg));
	  write_config++;
	  break;
	case 'r':
	  actions = add_info(gres);
	  break;
	case 'R':
	  actions = add_action(sres, strdup(optarg));
	  write_config++;
	  break;
	case 'l':
	  actions = add_info(gloads);
	  break;
#ifdef XF86_DISABLE_MODULE
	case 'L':
	  actions = add_action(sloads, strdup(optarg));
	  write_config++;
	  break;
	case 'm':
	  actions = add_info(gdisable);
	  break;
	case 'M':
	  actions = add_action(sdisable, strdup(optarg));
	  write_config++;
	  break;
#endif
	case 'e':
	  actions = add_info(gextopts);
	  break;
	case 'E':
	  actions = add_action(sextopt, strdup(optarg));
	  write_config++;
	  break;
	case 'a':
	  actions = add_action(gextopt, strdup(optarg));
	  break;
	default:
	  usage (EXIT_FAILURE);
	}
    }

#define CONFPATH "%A,%R,/etc/X11/%R,%P/etc/X11/%R,%E,%F,/etc/X11/%F,"	\
    "%P/etc/X11/%F,/etc/X11/%X,/etc/%X,%P/etc/X11/%X.%H,"		\
    "%P/etc/X11/%X,%P/lib/X11/%X.%H,%P/lib/X11/%X"

  if (((argc - optind) > 2) ||
      ((argc - optind) == 0) ||
      (write_config && ((argc - optind) == 1)))
    usage(EXIT_FAILURE);

  oldconfig = argv[optind];
  newconfig = (optind+1<argc)?argv[optind+1]:"./xorg2.conf";

  // fix strange SEGV on empty file names
  if (oldconfig[0]==0) exit (EXIT_FAILURE);

  if (!(filename = xf86openConfigFile (CONFPATH, oldconfig, NULL)))
      exit (EXIT_FAILURE);

  if ((conf = xf86readConfigFile ()) == NULL)
    {
      fprintf (stderr, "Problem when parsing config file\n");
      exit (EXIT_FAILURE);
    }
  xf86closeConfigFile ();

  for (; actions != NULL; actions = actions->list.next) {
  switch (actions->type)
    {
    case gdevice:
      {
	XF86ConfDevicePtr p = find_device(conf);
	if (p && xf86nameCompare(p->dev_identifier,"UnknownDevice"))
	  printf("%s\n",p->dev_driver);
      }
      break;
    case sdevice:
      setup_device(conf, actions->arg);
      break;
    case gmon:
      {
	XF86ConfMonitorPtr p = find_monitor(conf);
	if (p && xf86nameCompare(p->mon_identifier,"UnknownMonitor"))
	  printf("%s\n",p->mon_identifier);

      }
      break;
    case smon:
      setup_monitor(conf, actions->arg);
      break;
    case gcolor:
      printf("%d\n",find_screen(conf)->scrn_defaultdepth);
      break;
    case scolor:
      {
	XF86ConfScreenPtr screen;
	for (screen = find_screen(conf); screen; screen = screen->list.next)
	  screen->scrn_defaultdepth = atoi(actions->arg);
      }
      break;
    case gres:
      //prints first available mode in the first display subsection
      {
	XF86ConfScreenPtr screen = find_screen(conf);
	if (screen &&
	    screen->scrn_display_lst &&
	    screen->scrn_display_lst->disp_mode_lst &&
	    screen->scrn_display_lst->disp_mode_lst->mode_name)
	  printf("%s\n",
		 find_screen(conf)->
		 scrn_display_lst->
		 disp_mode_lst->mode_name);
	break;
      }
    case sres:
      setup_resolution(conf, actions->arg);
      break;
    case gloads:
      get_loads(conf);
      break;
#ifdef XF86_DISABLE_MODULE
    case sloads:
      set_loads(conf, actions->arg);
      break;
    case gdisable:
      get_disabled(conf);
      break;
    case sdisable:
      set_disable(conf, actions->arg);
      break;
#endif
    case gextopts:
      get_extopts(conf);
      break;
    case sextopt:
      set_extopt(conf, actions->arg);
      break;
    case gextopt:
      get_extopt(conf, actions->arg);
      break;
    default:
      fprintf(stderr,"fatal:unknown action type\n");
      exit(EXIT_FAILURE);
    }
    }

  if (write_config)
    xf86writeConfigFile (newconfig, conf);

  return EXIT_SUCCESS;
}
